import json
import logging
import os
import sqlite3
import tempfile
import uuid
from datetime import datetime, timezone
from decimal import Decimal
from pathlib import Path

from django.conf import settings
from django.core.mail import send_mail
from django.db import connections
from django.template.loader import render_to_string
from django.utils.text import slugify
from django_rq import job
from rq import get_current_job

from arkindex.documents.models import CorpusExport, CorpusExportState

BASE_DIR = Path(__file__).absolute().parent

# Maximum number of rows that will be stored in memory at once before being sent to SQLite.
BATCH_SIZE = 10000

logger = logging.getLogger(__name__)

# Map SQLite table names to PostgreSQL query files
EXPORT_QUERIES = [
    "image_server",
    "image",
    "worker_version",
    "worker_run",
    "element",
    "element_path",
    "transcription",
    "classification",
    "entity_type",
    "entity",
    "transcription_entity",
    "metadata",
    "dataset",
    "dataset_element",
]


def run_pg_query(query, source_db):
    """
    Run a single Postgresql query and split the results into chunks.
    When a name is given to a cursor, psycopg2 uses a server-side cursor; we just use a random string as a name.
    """
    db = connections[source_db]

    # Make sure a connection is open and available for export databases
    if source_db != "default" and db.connection is None:
        db.connect()

    with db.create_cursor(name=str(uuid.uuid4())) as pg_cursor:
        pg_cursor.itersize = BATCH_SIZE
        pg_cursor.execute(query)

        while True:
            out = pg_cursor.fetchmany(BATCH_SIZE)
            if not out:
                return
            yield out


def save_sqlite(rows, table, cursor):
    """
    Write a chunk of rows into an SQLite table
    """
    def _serialize(value):
        # Serialize UUID as string
        if isinstance(value, uuid.UUID):
            return str(value)

        # Serialize list and dicts as json
        if isinstance(value, (list, dict)):
            return json.dumps(value)

        # Serialize Decimal numbers as regular floats
        # We don't care about any loss of precision because SQLite will loose that precision anyway
        if isinstance(value, Decimal):
            return float(value)

        # Show very explicit error messages if we stumble upon an unexpected type
        # https://docs.python.org/3.10/library/sqlite3.html#sqlite-and-python-types
        assert value is None or isinstance(value, (int, float, str, bytes)), f"Type {type(value)} is not supported by sqlite3"

        return value

    rows = [
        list(map(_serialize, row))
        for row in rows
    ]

    # Build the parameterized query by counting the columns in a CSV row and repeating '?' parameters
    insert_args = ",".join("?" for _ in range(len(rows[0])))
    query = f"INSERT INTO {table} VALUES ({insert_args})"
    cursor.executemany(query, rows)


def send_email(subject, template_name, corpus_export, **context):
    context["corpus_export"] = corpus_export
    context["hostname"] = settings.PUBLIC_HOSTNAME
    sent = send_mail(
        subject=subject,
        message=render_to_string(
            template_name,
            context=context,
        ),
        from_email=None,
        recipient_list=[corpus_export.user.email],
        fail_silently=True,
    )
    if sent == 0:
        logger.error(f"Failed to send email to {corpus_export.user.email}")


@job("export", timeout=settings.RQ_TIMEOUTS["export_corpus"])
def export_corpus(corpus_export: CorpusExport) -> None:
    _, db_path = tempfile.mkstemp(suffix=".db")
    try:
        rq_job = get_current_job()
        corpus_export.state = CorpusExportState.Running
        corpus_export.save()

        export_source = f"{corpus_export.corpus_id}"
        if corpus_export.source != "default":
            export_source += f" from source {corpus_export.source}"

        logger.info(f"Exporting corpus {export_source} into {db_path}")
        db = sqlite3.connect(db_path)
        cursor = db.cursor()

        # Initialize all the tables
        cursor.executescript((BASE_DIR / "structure.sql").read_text())

        for i, name in enumerate(EXPORT_QUERIES):
            query = (BASE_DIR / f"{name}.sql").read_text()
            logger.info(f"Running query {i+1}/{len(EXPORT_QUERIES)} {name}")

            if rq_job:
                rq_job.set_progress(i / (len(EXPORT_QUERIES) + 1))
            for chunk in run_pg_query(query.format(corpus_id=corpus_export.corpus_id), corpus_export.source):
                save_sqlite(chunk, name, cursor)

            db.commit()

        logger.info("Creating indexes")
        if rq_job:
            rq_job.set_progress(len(EXPORT_QUERIES) / (len(EXPORT_QUERIES) + 1))
        cursor.executescript((BASE_DIR / "indexes.sql").read_text())

        db.close()

        # Give a nice feedback on the upload portion of the export using boto3's upload callback
        # This can be useful for large exports, when the database size is measured in gigabytes
        size = os.stat(db_path).st_size
        uploaded = 0

        def progress_callback(bytes_transferred):
            nonlocal uploaded
            uploaded += bytes_transferred
            if rq_job:
                rq_job.set_progress(uploaded / size)

        if rq_job:
            rq_job.set_description(f"Uploading export for corpus {corpus_export.corpus.name}")
            rq_job.set_progress(0.0)

        # Set the filename in a HTTP header to make downloaded exports nicer
        filename = f'{slugify(corpus_export.corpus.name)[:100]}-{datetime.now(timezone.utc).strftime("%Y%m%d-%H%M%S")}.sqlite'
        corpus_export.s3_object.upload_file(
            db_path,
            Callback=progress_callback,
            ExtraArgs={
                "ContentType": "application/vnd.sqlite3",
                "ContentDisposition": f'attachment; filename="{filename}"'
            },
        )

        corpus_export.state = CorpusExportState.Done
        corpus_export.save()

        send_email(
            "Arkindex project export completed",
            "export_success.html",
            corpus_export,
        )
    except Exception as e:
        corpus_export.state = CorpusExportState.Failed
        corpus_export.save()
        send_email(
            "Arkindex project export failed",
            "export_error.html",
            corpus_export,
            error=e.args[0],
        )
        raise
    finally:
        os.unlink(db_path)
