import itertools

from django.db import connections
from django.db.backends.base.base import _thread

from arkindex.documents.models import Element, ElementPath
from arkindex.project.tests import FixtureTestCase
from arkindex.project.tools import build_tree


class TestEditElementPath(FixtureTestCase):
    """
    Test ElementPath editing algorithms
    """

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.element_type = cls.corpus.types.get(slug="volume")

    def check_parents(self, elements, element_name, *paths):
        """
        Check an element's parent paths by names
        """
        tree = {
            elt_id: [e.path for e in elts]
            for elt_id, elts in itertools.groupby(
                ElementPath.objects.all().order_by("element_id"),
                lambda p: p.element_id,
            )
        }

        if paths is None or len(paths) < 1:
            self.assertNotIn(elements[element_name].id, tree)
            return

        element_names = {e.id: e.name for e in elements.values()}

        expected_paths = [
            [elements[parent_name].id for parent_name in path]
            for path in paths
        ]
        actual_paths = tree.get(elements[element_name].id, [])
        readable_actual_paths = [
            [element_names[id] for id in path]
            for path in actual_paths
        ]

        return self.assertCountEqual(
            actual_paths,
            expected_paths,
            f"Expected paths {paths} in element {element_name}, got {readable_actual_paths}"
        )

    def test_add_first_parent(self):
        r"""
        Test adding A as a parent of B.
            X   Y
             \ /
              A
             + \
            B   C
           / \   \
          D   E   F
         /
        G
        """
        elements = build_tree(
            {
                "A": ["X", "Y"],
                "C": "A",
                "F": "C",
                "D": "B",
                "E": "B",
                "G": "D",
            },
            corpus=self.corpus,
            type=self.element_type,
        )

        self.check_parents(elements, "X", [])
        self.check_parents(elements, "Y", [])
        self.check_parents(elements, "A", ["X"],
                                          ["Y"])
        self.check_parents(elements, "B", [])
        self.check_parents(elements, "C", ["X", "A"],
                                          ["Y", "A"])
        self.check_parents(elements, "D", ["B"])
        self.check_parents(elements, "E", ["B"])
        self.check_parents(elements, "F", ["X", "A", "C"],
                                          ["Y", "A", "C"])
        self.check_parents(elements, "G", ["B", "D"])

        with self.assertExactQueries(
            "add_first_parent.sql", params={
                # add_parent uses transaction.atomic(), and we are running in a unit test, which is already in a transaction.
                # This will cause a savepoint to be created, with a name that is hard to mock.
                "savepoint": f"s{_thread.get_ident()}_x{connections['default'].savepoint_state + 1}",
                "A": elements["A"].id,
                "B": elements["B"].id,
                # Element A has two parents so it has two paths.
                # add_parent will pick the first one to perform updates on the new child's paths,
                # so it will be seen in the SQL queries. To avoid intermittent failures,
                # add_parent sorts parent paths by `path`, so we apply the same sort here.
                # The paths only contain one ID, X's or Y's.
                "first_parent": elements["A"].paths.order_by("path").first().path[0],
            }
        ):
            elements["B"].add_parent(elements["A"])

        self.check_parents(elements, "X", [])
        self.check_parents(elements, "Y", [])
        self.check_parents(elements, "A", ["X"],
                                          ["Y"])
        self.check_parents(elements, "B", ["X", "A"],
                                          ["Y", "A"])
        self.check_parents(elements, "C", ["X", "A"],
                                          ["Y", "A"])
        self.check_parents(elements, "D", ["X", "A", "B"],
                                          ["Y", "A", "B"])
        self.check_parents(elements, "E", ["X", "A", "B"],
                                          ["Y", "A", "B"])
        self.check_parents(elements, "F", ["X", "A", "C"],
                                          ["Y", "A", "C"])
        self.check_parents(elements, "G", ["X", "A", "B", "D"],
                                          ["Y", "A", "B", "D"])

    def test_add_second_parent(self):
        r"""
        Test adding A as a parent of B.
            X   Y
             \ /
          K   A
           \ + \
            B   C
           / \   \
          D   E   F
         /
        G
        """
        elements = build_tree(
            {
                "A": ["X", "Y"],
                "C": "A",
                "F": "C",
                "B": "K",
                "D": "B",
                "E": "B",
                "G": "D",
            },
            corpus=self.corpus,
            type=self.element_type,
        )

        self.check_parents(elements, "X", [])
        self.check_parents(elements, "Y", [])
        self.check_parents(elements, "K", [])
        self.check_parents(elements, "A", ["X"],
                                          ["Y"])
        self.check_parents(elements, "B", ["K"])
        self.check_parents(elements, "C", ["X", "A"],
                                          ["Y", "A"])
        self.check_parents(elements, "D", ["K", "B"])
        self.check_parents(elements, "E", ["K", "B"])
        self.check_parents(elements, "F", ["X", "A", "C"],
                                          ["Y", "A", "C"])
        self.check_parents(elements, "G", ["K", "B", "D"])

        with self.assertExactQueries(
            "add_second_parent.sql", params={
                # add_parent uses transaction.atomic(), and we are running in a unit test, which is already in a transaction.
                # This will cause a savepoint to be created, with a name that is hard to mock.
                "savepoint": f"s{_thread.get_ident()}_x{connections['default'].savepoint_state + 1}",
                "A": elements["A"].id,
                "B": elements["B"].id,
                "K": elements["K"].id,
            }
        ):
            elements["B"].add_parent(elements["A"])

        self.check_parents(elements, "X", [])
        self.check_parents(elements, "Y", [])
        self.check_parents(elements, "K", [])
        self.check_parents(elements, "A", ["X"],
                                          ["Y"])
        self.check_parents(elements, "B", ["X", "A"],
                                          ["Y", "A"],
                                          ["K"])
        self.check_parents(elements, "C", ["X", "A"],
                                          ["Y", "A"])
        self.check_parents(elements, "D", ["X", "A", "B"],
                                          ["Y", "A", "B"],
                                          ["K", "B"])
        self.check_parents(elements, "E", ["X", "A", "B"],
                                          ["Y", "A", "B"],
                                          ["K", "B"])
        self.check_parents(elements, "F", ["X", "A", "C"],
                                          ["Y", "A", "C"])
        self.check_parents(elements, "G", ["X", "A", "B", "D"],
                                          ["Y", "A", "B", "D"],
                                          ["K", "B", "D"])

    def test_add_parent_cycle(self):
        """
        Check we can't create cycles with add_parent
        A <-- B <-- C (parent <-- child)
        Tries to add C as a parent of A.
        """
        elements = build_tree(
            {
                "C": "B",
                "B": "A",
            },
            corpus=self.corpus,
            type=self.element_type,
        )
        with self.assertRaises(ValueError):
            elements["A"].add_parent(elements["C"])

    def test_add_parent_empty(self):
        """
        Check adding a parent works with an empty tree
        """
        elements = build_tree({"A": [], "B": []}, corpus=self.corpus, type=self.element_type)

        self.check_parents(elements, "A", [])
        self.check_parents(elements, "B", [])

        elements["B"].add_parent(elements["A"])

        self.check_parents(elements, "A", [])
        self.check_parents(elements, "B", ["A"])

    def test_remove_child(self):
        r"""
        Test removing child B from parent A.
            X   Y
             \ /
          K   A
           \ - \
            B   C
           / \   \
          D   E   F
         /
        G
        """
        elements = build_tree(
            {
                "A": ["X", "Y"],
                "C": "A",
                "F": "C",
                "B": ["A", "K"],
                "D": "B",
                "E": "B",
                "G": "D",
            },
            corpus=self.corpus,
            type=self.element_type,
        )

        self.check_parents(elements, "X", [])
        self.check_parents(elements, "Y", [])
        self.check_parents(elements, "K", [])
        self.check_parents(elements, "A", ["X"],
                                          ["Y"])
        self.check_parents(elements, "B", ["X", "A"],
                                          ["Y", "A"],
                                          ["K"])
        self.check_parents(elements, "C", ["X", "A"],
                                          ["Y", "A"])
        self.check_parents(elements, "D", ["X", "A", "B"],
                                          ["Y", "A", "B"],
                                          ["K", "B"])
        self.check_parents(elements, "E", ["X", "A", "B"],
                                          ["Y", "A", "B"],
                                          ["K", "B"])
        self.check_parents(elements, "F", ["X", "A", "C"],
                                          ["Y", "A", "C"])
        self.check_parents(elements, "G", ["X", "A", "B", "D"],
                                          ["Y", "A", "B", "D"],
                                          ["K", "B", "D"])

        with self.assertExactQueries(
            "remove_child.sql", params={
                # remove_child uses transaction.atomic(), and we are running in a unit test, which is already in a transaction.
                # This will cause a savepoint to be created, with a name that is hard to mock.
                "savepoint": f"s{_thread.get_ident()}_x{connections['default'].savepoint_state + 1}",
                "A": elements["A"].id,
                "B": elements["B"].id,
            }
        ):
            elements["A"].remove_child(elements["B"])

        self.check_parents(elements, "X", [])
        self.check_parents(elements, "Y", [])
        self.check_parents(elements, "K", [])
        self.check_parents(elements, "A", ["X"],
                                          ["Y"])
        self.check_parents(elements, "B", ["K"])
        self.check_parents(elements, "C", ["X", "A"],
                                          ["Y", "A"])
        self.check_parents(elements, "D", ["K", "B"])
        self.check_parents(elements, "E", ["K", "B"])
        self.check_parents(elements, "F", ["X", "A", "C"],
                                          ["Y", "A", "C"])
        self.check_parents(elements, "G", ["K", "B", "D"])

    def test_remove_child_update(self):
        r"""
        Scenario where updates are required
            X   Y
             \ /
              A
             - \
            B   C
           / \
          D   E
               \
                F
        """

        elements = build_tree(
            {
                "A": ["X", "Y"],
                "B": "A",
                "C": "A",
                "D": "B",
                "E": "B",
                "F": "E",
            },
            corpus=self.corpus,
            type=self.element_type,
        )

        self.check_parents(elements, "X", [])
        self.check_parents(elements, "Y", [])
        self.check_parents(elements, "A", ["X"],
                                          ["Y"])
        self.check_parents(elements, "B", ["X", "A"],
                                          ["Y", "A"])
        self.check_parents(elements, "C", ["X", "A"],
                                          ["Y", "A"])
        self.check_parents(elements, "D", ["X", "A", "B"],
                                          ["Y", "A", "B"])
        self.check_parents(elements, "E", ["X", "A", "B"],
                                          ["Y", "A", "B"])
        self.check_parents(elements, "F", ["X", "A", "B", "E"],
                                          ["Y", "A", "B", "E"])

        path1, path2 = elements["B"].paths.filter(path__last=elements["A"].id)

        # remove_child picks the first parent path to perform an update on it, but it really just needs any random one,
        # therefore it is retrieved without an ordering. This can cause intermittent failures in tests since either
        # X or Y will be picked. Adding an ORDER BY in the original method would not help much either since we cannot
        # order by any value that would not change randomly during the execution.
        # This class implements a workaround that works with assertExactQueries: the __str__ will be called after
        # remove_child has run, and just before assertExactQueries checks the queries against the expected SQL.
        # B will only have one remaining path: the first path that got picked by remove_child for the update.
        # The other path will have been deleted. We can therefore get this remaining path and compare it by its ID
        # to the two paths that we had before, and pick the parent that was in the old version of this path.
        class FirstParent:
            def __str__(self):
                path_id = elements["B"].paths.get().id
                if path1.id == path_id:
                    return str(path1.path[0])
                if path2.id == path_id:
                    return str(path2.path[0])
                raise AssertionError("Unexpected top-level path ID")

        with self.assertExactQueries(
            "remove_child_last_parent.sql", params={
                # remove_child uses transaction.atomic(), and we are running in a unit test, which is already in a transaction.
                # This will cause a savepoint to be created, with a name that is hard to mock.
                "savepoint": f"s{_thread.get_ident()}_x{connections['default'].savepoint_state + 1}",
                "A": elements["A"].id,
                "B": elements["B"].id,
                "first_parent": FirstParent(),
            }
        ):
            elements["A"].remove_child(elements["B"])

        self.check_parents(elements, "X", [])
        self.check_parents(elements, "Y", [])
        self.check_parents(elements, "B", [])
        self.check_parents(elements, "A", ["X"],
                                          ["Y"])
        self.check_parents(elements, "C", ["X", "A"],
                                          ["Y", "A"])
        self.check_parents(elements, "D", ["B"])
        self.check_parents(elements, "E", ["B"])
        self.check_parents(elements, "F", ["B", "E"])

    def test_delete_element(self):
        r"""
        Test the signal works properly and is able to remove all paths when removing an Element.
            X   Y
             \ /
              A
             / \
           [B]  C
           / \
          D   E
               \
                F
        """

        elements = build_tree(
            {
                "A": ["X", "Y"],
                "B": "A",
                "C": "A",
                "D": "B",
                "E": "B",
                "F": "E",
            },
            corpus=self.corpus,
            type=self.element_type,
        )

        self.check_parents(elements, "X", [])
        self.check_parents(elements, "Y", [])
        self.check_parents(elements, "A", ["X"],
                                          ["Y"])
        self.check_parents(elements, "B", ["X", "A"],
                                          ["Y", "A"])
        self.check_parents(elements, "C", ["X", "A"],
                                          ["Y", "A"])
        self.check_parents(elements, "D", ["X", "A", "B"],
                                          ["Y", "A", "B"])
        self.check_parents(elements, "E", ["X", "A", "B"],
                                          ["Y", "A", "B"])
        self.check_parents(elements, "F", ["X", "A", "B", "E"],
                                          ["Y", "A", "B", "E"])

        elements["B"].delete()

        # Check parents still exists
        for parent in ("X", "Y", "A", "C"):
            elements[parent].refresh_from_db()
            self.assertIsNotNone(elements[parent].id)

        self.check_parents(elements, "X", [])
        self.check_parents(elements, "Y", [])
        self.check_parents(elements, "A", ["X"],
                                          ["Y"])
        self.check_parents(elements, "C", ["X", "A"],
                                          ["Y", "A"])
        # Check all children are deleted
        for child in ("B", "D", "E", "F"):
            with self.assertRaises(Element.DoesNotExist):
                elements[child].refresh_from_db()

    def test_add_many_parents(self):
        """
        Ensure that add_parent still works when adding a lot of parents.
        We previously had issues that were not detected in unit tests because we were not testing with enough parents!
        """
        parent_names = [f"Parent {i}" for i in range(50)]
        elements = build_tree(
            {
                "Child 1": ["Element"],
                "Child 2": ["Element"],
                "Element": parent_names,
            },
            corpus=self.corpus,
            type=self.element_type,
        )

        self.check_parents(elements, "Element", *[[parent_name] for parent_name in parent_names])
        self.check_parents(elements, "Child 1", *[[parent_name, "Element"] for parent_name in parent_names])
        self.check_parents(elements, "Child 2", *[[parent_name, "Element"] for parent_name in parent_names])
