import shlex
from collections import defaultdict
from datetime import timedelta
from functools import wraps
from os import path
from typing import Dict, List, Sequence, Tuple
from uuid import UUID

from django.conf import settings
from django.db.models import Prefetch, prefetch_related_objects
from django.utils import timezone
from django.utils.functional import cached_property
from rest_framework.exceptions import ValidationError

from arkindex.images.models import ImageServer
from arkindex.ponos.models import GPU, Task, task_token_default


class ProcessBuilder:

    def __init__(self, process) -> None:
        self.process = process
        self.run = process.get_last_run() + 1
        self.mode_key = process.mode.value
        # List of tasks instances for this process (not yet saved)
        self.tasks = []
        # Mapping between a task's slug and the slugs of their direct parents
        self.tasks_parents = defaultdict(list)

    def prefetch_worker_runs(func):
        @wraps(func)
        def _wraps(self, *args, **kwargs):
            from arkindex.process.models import WorkerRun

            prefetch_cache = getattr(self.process, "_prefetched_objects_cache", {})
            if "worker_runs" not in prefetch_cache.keys():
                prefetch_related_objects(
                    [self.process],
                    Prefetch(
                        "worker_runs",
                        queryset=(
                            WorkerRun.objects
                            .using("default")
                            .select_related("version__worker__repository", "model_version__model")
                        ),
                    )
                )
            func(self, *args, **kwargs)

        return _wraps

    def _get_task_depth(self, slug) -> int:
        """
        Determines the depth of a task within a tree without making any database queries
        """
        parent_slugs = self.tasks_parents[slug]
        if not parent_slugs:
            return 0

        return max(
            self._get_task_depth(parent_slug)
            for parent_slug in parent_slugs
        ) + 1

    def _build_task(
        self,
        command,
        slug,
        *,
        env={},
        image=None,
        artifact=None,
        extra_files={},
        requires_gpu=False,
        shm_size=None,
        worker_run=None,
    ) -> None:
        """
        Build a Task with default attributes and add it to the current stack.
        Depth is not set while building individual Task instances.
        """
        if image is None:
            image = settings.ARKINDEX_TASKS_IMAGE
        token = task_token_default()
        self.tasks.append(
            Task(
                process=self.process,
                command=command,
                slug=slug,
                token=token,
                env={**env, "ARKINDEX_TASK_TOKEN": token},
                run=self.run,
                image=image,
                requires_gpu=requires_gpu,
                shm_size=shm_size,
                extra_files=extra_files,
                image_artifact_id=artifact,
                worker_run=worker_run,
            )
        )

    def _create_worker_versions_cache(
        self,
        worker_versions_attrs: Sequence[Tuple[UUID, UUID | None, UUID | None]],
    ) -> None:
        """
        Cache worker versions that generate results on a corpus.
        Each item must be a tuple with a WorkerVersion ID, a nullable
        ModelVersion ID and a nullable Configuration ID.
        """
        from arkindex.process.models import CorpusWorkerVersion

        assert self.process.corpus_id is not None, "Worker versions cache requires the process to be linked to a corpus"

        CorpusWorkerVersion.objects.bulk_create([
            CorpusWorkerVersion(
                corpus_id=self.process.corpus_id,
                worker_version_id=worker_version_id,
                model_version_id=model_version_id,
                worker_configuration_id=conf_id,
            )
            for worker_version_id, model_version_id, conf_id in worker_versions_attrs
        ], ignore_conflicts=True)

    def _create_fake_worker_run(self, worker_version):
        """
        Returns a saved WorkerRun instance
        """
        run, _ = self.process.worker_runs.get_or_create(version=worker_version, parents=[])
        return run

    def _get_elements_json_chunks(self) -> List[str]:
        """
        Elements can be distributed by chunks by the initialisation task
        """
        if (chunks := self.process.chunks) <= 1:
            return ["elements.json"]
        return [f"elements_chunk_{n}.json" for n in range(1, chunks + 1)]

    @property
    def base_env(self) -> Dict[str, str]:
        env = settings.PONOS_DEFAULT_ENV.copy()
        env["ARKINDEX_PROCESS_ID"] = str(self.process.id)
        if self.process.corpus_id:
            env["ARKINDEX_CORPUS_ID"] = str(self.process.corpus_id)
        return env

    @cached_property
    def active_gpu_agents(self) -> bool:
        return GPU.objects.filter(agent__last_ping__gt=timezone.now() - timedelta(seconds=30)).exists()

    @prefetch_worker_runs
    def validate_gpu_requirement(self):
        from arkindex.process.models import FeatureUsage

        if (
            not self.process.use_gpu
            and any(run.version.gpu_usage == FeatureUsage.Required for run in self.process.worker_runs.all())
        ):
            raise ValidationError("Some worker versions require a GPU and the `use_gpu` option is disabled.")

    @prefetch_worker_runs
    def validate_archived(self):
        if any(run.version.worker.archived for run in self.process.worker_runs.all()):
            raise ValidationError("Some worker versions are on archived workers and cannot be executed.")

        if any(
            run.model_version.model.archived
            for run in self.process.worker_runs.all()
            if run.model_version is not None
        ):
            raise ValidationError("Some model versions are on archived models and cannot be executed.")

    def validate_s3(self) -> None:
        if not self.process.bucket_name:
            raise ValidationError("Missing S3 bucket name")
        if self.process.folder_type is None:
            raise ValidationError("Missing folder type")
        if self.process.element_type is None:
            raise ValidationError("Missing element type")

    def validate_workers(self) -> None:
        self.validate_gpu_requirement()
        self.validate_archived()

    def validate_dataset(self) -> None:
        self.validate_gpu_requirement()
        self.validate_archived()

    def validate(self) -> None:
        """
        Ensure the process can be built.
        Extra validation can be executed depending on the process' type.
        """
        from arkindex.process.models import ActivityState, ProcessMode

        if self.run > 0 and not self.process.is_final:
            raise ValidationError("Cannot run a process that is not in a final state.")
        if self.process.mode in (ProcessMode.Local, ProcessMode.Template):
            raise ValidationError(f"Cannot run a process with mode {self.process.mode}")
        if (
            self.process.mode != ProcessMode.Workers
            and (
                self.process.use_cache
                or self.process.activity_state != ActivityState.Disabled
            )
        ):
            raise ValidationError("Only worker processes can be started with cache or worker activities")
        if extra_validation := getattr(self, f"validate_{self.mode_key}", None):
            extra_validation()

    def build_files(self):
        from arkindex.process.models import WorkerVersion

        worker_run = self._create_fake_worker_run(worker_version=WorkerVersion.objects.imports_version)

        self._build_task(
            command=f"python -m arkindex_tasks.import_files {self.process.id}",
            slug="import_files",
            env={
                **self.base_env,
                "ARKINDEX_WORKER_RUN_ID": str(worker_run.id),
            },
            worker_run=worker_run,
        )
        self._create_worker_versions_cache([(settings.IMPORTS_WORKER_VERSION, None, None)])

    def build_iiif(self):
        from arkindex.process.models import WorkerVersion

        worker_run = self._create_fake_worker_run(worker_version=WorkerVersion.objects.imports_version)
        self._build_task(
            command=f"python -m arkindex_tasks.import_iiif.process {self.process.id}",
            slug="import_iiif",
            env={
                **self.base_env,
                "ARKINDEX_WORKER_RUN_ID": str(worker_run.id),
            },
            worker_run=worker_run,
        )
        self._create_worker_versions_cache([(settings.IMPORTS_WORKER_VERSION, None, None)])

    @prefetch_worker_runs
    def build_workers(self):
        # Build the initialisation task listing elements by chunks
        args = [
            "python", "-m", "arkindex_tasks.init_elements", str(self.process.id),
            "--chunks-number", str(self.process.chunks),
        ]
        if self.process.use_cache:
            args.append("--use-cache")
        import_task_slug = "initialisation"
        self._build_task(
            command=" ".join(args),
            slug=import_task_slug,
            env=self.base_env,
        )

        # Distribute worker run tasks
        worker_runs = list(self.process.worker_runs.all())
        chunks = self._get_elements_json_chunks()
        for index, chunk in enumerate(chunks, start=1):
            elements_path = shlex.quote(path.join("/data", import_task_slug, chunk))
            for worker_run in worker_runs:
                task, parent_slugs = worker_run.build_task(
                    self.process,
                    self.base_env,
                    import_task_slug,
                    elements_path,
                    chunk=index if len(chunks) > 1 else None,
                    workflow_runs=worker_runs,
                    run=self.run,
                    active_gpu_agents=self.active_gpu_agents,
                )
                self.tasks.append(task)
                self.tasks_parents[task.slug].extend(parent_slugs)

        # Populate the ML results cache table
        self._create_worker_versions_cache([
            (
                run.version_id,
                run.model_version_id,
                run.configuration_id,
            ) for run in worker_runs
        ])

    def build_s3(self):
        from arkindex.process.models import WorkerVersion

        env = {**self.base_env}
        env["INGEST_S3_ACCESS_KEY"] = settings.INGEST_S3_ACCESS_KEY
        env["INGEST_S3_SECRET_KEY"] = settings.INGEST_S3_SECRET_KEY
        if settings.INGEST_S3_ENDPOINT:
            env["INGEST_S3_ENDPOINT"] = settings.INGEST_S3_ENDPOINT
        if settings.INGEST_S3_REGION:
            env["INGEST_S3_REGION"] = settings.INGEST_S3_REGION

        worker_run = self._create_fake_worker_run(worker_version=WorkerVersion.objects.imports_version)
        env["ARKINDEX_WORKER_RUN_ID"] = str(worker_run.id)

        command = (
            "python -m arkindex_tasks.import_s3"
            f" --corpus={self.process.corpus_id}"
            f" --bucket={shlex.quote(self.process.bucket_name)}"
            f" --folder-type={shlex.quote(self.process.folder_type.slug)}"
            f" --page-type={shlex.quote(self.process.element_type.slug)}"
            f" --iiif-base-url={shlex.quote(ImageServer.objects.ingest.url)}"
        )
        if settings.INGEST_PREFIX_BY_BUCKET_NAME:
            command += " --bucket-prefix"
        if self.process.prefix:
            command += f" --prefix={shlex.quote(self.process.prefix)}"
        if self.process.element:
            command += f" --element={self.process.element_id}"
        self._build_task(
            command=command,
            slug="import_s3",
            env=env,
        )
        self._create_worker_versions_cache([(settings.IMPORTS_WORKER_VERSION, None, None)])

    @prefetch_worker_runs
    def build_dataset(self):
        worker_runs = list(self.process.worker_runs.all())
        chunks = self._get_elements_json_chunks()
        for index, chunk in enumerate(chunks, start=1):
            for worker_run in worker_runs:
                task, parent_slugs = worker_run.build_task(
                    self.process,
                    self.base_env,
                    None,
                    None,
                    chunk=index if len(chunks) > 1 else None,
                    workflow_runs=worker_runs,
                    run=self.run,
                )
                self.tasks.append(task)
                self.tasks_parents[task.slug].extend(parent_slugs)

    def build(self, *args, **kwargs):
        if not (_build := getattr(self, f"build_{self.mode_key}", None)):
            raise NotImplementedError
        _build(*args, **kwargs)

    def save(self):
        if len(self.tasks) == 0:
            raise ValidationError("The process has no task to be saved")

        # Compute the depth of each task once the whole graph is built and before saving
        for task in self.tasks:
            task.depth = self._get_task_depth(task.slug)

        # Build all tasks and task relationships in two queries
        Task.objects.bulk_create(self.tasks)

        # Use a tasks by slug map to find tasks more easily
        tasks = {task.slug: task for task in self.tasks}

        Task.parents.through.objects.bulk_create(
            Task.parents.through(
                from_task=tasks[child_slug],
                to_task=tasks[parent_slug],
            )
            for child_slug, parent_slugs in self.tasks_parents.items()
            for parent_slug in parent_slugs
        )
        return tasks
