from django.conf import settings
from django.db.models import Q
from django.shortcuts import get_object_or_404
from django.views.decorators.cache import cache_page
from drf_spectacular.utils import extend_schema, extend_schema_view
from rest_framework import status
from rest_framework.exceptions import APIException, PermissionDenied
from rest_framework.serializers import CharField, Serializer

from arkindex.documents.models import Corpus
from arkindex.process.models import Process, Repository
from arkindex.users.models import Role
from arkindex.users.utils import filter_rights, get_max_level, has_access


class ACLMixin:
    """
    Access control mixin using the generic Right table.
    """
    _user = None

    def __init__(self, *args, user=None, **kwargs):
        super().__init__(*args, **kwargs)
        self._user = user

    @property
    def user(self):
        return self._user or self.request.user

    def has_access(self, instance, level, skip_public=False):
        return has_access(self.user, instance, level, skip_public=skip_public)


class RepositoryACLMixin(ACLMixin):

    @property
    def readable_repositories(self):
        return Repository.objects.filter(
            id__in=filter_rights(self.user, Repository, Role.Guest.value).values("id")
        )

    @property
    def executable_repositories(self):
        return Repository.objects.filter(
            id__in=filter_rights(self.user, Repository, Role.Contributor.value).values("id")
        )

    def has_read_access(self, repo):
        return self.has_access(repo, Role.Guest.value)

    def has_execution_access(self, repo):
        return self.has_access(repo, Role.Contributor.value)

    def has_admin_access(self, repo):
        return self.has_access(repo, Role.Admin.value)


class WorkerACLMixin(ACLMixin):
    """
    Access control mixin for workers
    A public worker is considered as executable (i.e. any user has a contributor access)
    """

    def has_worker_access(self, worker, level):
        if worker.public and level <= Role.Contributor.value:
            return True
        if self.has_access(worker, level):
            return True
        if worker.repository is None:
            return False
        # User may be granted access to the worker via a right on its repository
        return self.has_access(worker.repository, level)

    def has_read_access(self, worker):
        return self.has_worker_access(worker, Role.Guest.value)

    def has_execution_access(self, worker):
        return self.has_worker_access(worker, Role.Contributor.value)

    def has_admin_access(self, worker):
        return self.has_worker_access(worker, Role.Admin.value)


class CorpusACLMixin(ACLMixin):

    def get_corpus(self, corpus_id, role=Role.Guest):
        """
        Try to retrieve a corpus according to user rights.
        Raises a 404 if the corpus does not exists and a 403 if the permissions are invalid
        """
        corpus = get_object_or_404(Corpus, id=corpus_id)
        if not self.has_access(corpus, role.value):
            raise PermissionDenied(detail=f"You do not have {str(role).lower()} access to this corpus.")
        return corpus

    def has_read_access(self, corpus):
        return self.has_access(corpus, Role.Guest.value)

    def has_write_access(self, corpus):
        return self.has_access(corpus, Role.Contributor.value)

    def has_admin_access(self, corpus):
        return self.has_access(corpus, Role.Admin.value)


class TrainingModelMixin(ACLMixin):
    """
    Access control mixin for machine learning models
    """

    def has_read_access(self, model):
        return self.has_access(model, Role.Guest.value)

    def has_write_access(self, model):
        return self.has_access(model, Role.Contributor.value)

    def has_admin_access(self, model):
        return self.has_access(model, Role.Admin.value)


class ProcessACLMixin(ACLMixin):
    """
    Rights on processes are determined depending on its nature.
    The process creator always has a read access.
    If the process is attached to a project, the project rights apply.
    """

    @property
    def readable_processes(self):
        # Return a queryset of accessible process
        return Process.objects.filter(
            Q(creator_id=self.user.id)
            | Q(corpus_id__in=filter_rights(self.user, Corpus, Role.Guest.value).values("id"))
        ).distinct()

    def process_access_level(self, process):
        if not process.corpus_id:
            return None
        # Return the access level on a single process
        return get_max_level(self.user, process.corpus)


class SelectionMixin:

    def get_selection(self, corpus_id=None):
        assert settings.ARKINDEX_FEATURES["selection"], "Selection feature is unavailable"
        queryset = self.request.user.selected_elements.all()
        if corpus_id is not None:
            queryset = queryset.filter(corpus_id=corpus_id)
        return queryset


class DeprecatedExceptionSerializer(Serializer):
    detail = CharField()


class ConflictAPIException(APIException):
    status_code = status.HTTP_409_CONFLICT
    default_code = "conflict"


class DeprecatedAPIException(APIException):
    status_code = 410
    default_detail = "This endpoint has been deprecated."
    default_code = "deprecated"


class DeprecatedMixin:
    # Add this mixin to an APIView to make it deprecated.

    serializer_class = DeprecatedExceptionSerializer
    pagination_class = None
    deprecation_message = None

    @classmethod
    def as_view(cls, *args, **kwargs):
        # A custom schema class cannot be assigned on the mixin as it would be linked to the
        # mixin itself, not a view class, so it would not override anything.
        # We therefore have to apply deprecation overrides on the view itself, and
        # we override as_view specifically as this is easier than writing a metaclass.
        extend_schema_view(**{
            method: extend_schema(
                deprecated=True,
                responses={DeprecatedAPIException.status_code: DeprecatedExceptionSerializer},
            )
            # Apply extend_schema only on methods that exist on this endpoint
            for method in cls.http_method_names
            if hasattr(cls, method)
        })(cls)
        return super().as_view(*args, **kwargs)

    def initial(self, request, *args, **kwargs):
        super().initial(request, *args, **kwargs)
        method = request.method.lower()
        if method in self.http_method_names and hasattr(self, method):
            raise DeprecatedAPIException(detail=getattr(self, "deprecation_message", None))


class CachedViewMixin:
    """
    Add this mixin to any class-based view to cache it.
    """
    cache_timeout = None
    cache_prefix = None

    def __init__(self, *args, **kwargs):
        if self.cache_timeout:
            self.dispatch = cache_page(
                self.cache_timeout,
                key_prefix=self.cache_prefix,
            )(self.dispatch)
