import re
from contextlib import contextmanager
from pathlib import Path
from typing import Any, Callable, Iterable, Mapping, Optional, Union

import sqlparse
from django.conf import settings
from django.contrib.contenttypes.models import ContentType
from django.db import DEFAULT_DB_ALIAS, connections
from django.test import TestCase
from django.test.utils import CaptureQueriesContext
from rest_framework.test import APITestCase

from arkindex.documents.models import Corpus
from arkindex.images.models import ImageServer
from arkindex.process.models import WorkerVersion
from arkindex.users.models import Group, User


class _AssertExactQueriesContext(CaptureQueriesContext):
    """
    Context manager for assertExactQueries.
    The implementation is inspired by assertNumQueries's own implementation.
    """

    def __init__(self, test_case, path, params, connection, skip=0):
        self.test_case = test_case
        self.path = settings.SQL_VALIDATION_DIR / Path(path)
        self.params = params
        self.skip = skip
        super().__init__(connection)

    def _format(self, query: str) -> str:
        """
        Format one or more SQL queries.

        Does not allow more than two newlines at once (at most one empty line).
        This still allows separating with empty lines in the expected SQL file,
        but prevents duplicate empty lines from appearing;
        sqlparse.format duplicates an empty line when it encounters one, causing assertion errors.
        """
        return re.sub(
            r"\n\n+",
            "\n\n",
            sqlparse.format(
                query,
                reindent=True,
                strip_comments=True,
                use_space_around_operators=True,
                indent_width=4,
            ).strip()
        )

    def __exit__(self, exc_type, exc_value, traceback):
        super().__exit__(exc_type, exc_value, traceback)
        if exc_type is not None:
            return

        # Django's logged queries are each on a single line without semicolons
        # so we lint them and assemble them to build a more readable diff for humans.
        actual_sql = self._format(
            ";".join(query["sql"] for query in self[self.skip:])
        )

        # When the file does not exist, try to create it with the current queries.
        # This avoids having to write all of the queries by hand when adding assertExactQueries for the first time.
        if not self.path.exists():
            try:
                self.path.write_text(actual_sql)
            except OSError as e:
                raise AssertionError(
                    f"Could not assert on SQL queries; the file at {self.path} did not exist. "
                    "A new file could not be created with the current SQL queries."
                ) from e
            else:
                raise AssertionError(
                    f"Could not assert on SQL queries; the file at {self.path} did not exist. "
                    "A file has been created with the current SQL queries, please check it manually.\n"
                    f"Parameters: {self.params!r}"
                )

        # Ignore newlines in the expected SQL, otherwise the reformatting might have duplicate line breaks
        expected_sql = self.path.read_text().strip()

        # Apply str.format on the SQL when needed to handle variable IDs
        if self.params:
            if isinstance(self.params, Mapping):
                expected_sql = expected_sql.format(**self.params)
            else:
                expected_sql = expected_sql.format(*self.params)

        expected_sql = self._format(expected_sql)

        self.test_case.maxDiff = None
        # Use the base TestCase's assertions, since they can already handle displaying a diff.
        self.test_case.assertEqual(expected_sql, actual_sql)


class FixtureMixin:
    """
    Add the database fixture to a test case
    """

    fixtures = ["data.json", ]

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.corpus = Corpus.objects.get(name="Unit Tests")
        cls.user = User.objects.get(email="user@user.fr")
        cls.group = Group.objects.get(name="User group")
        cls.superuser = User.objects.get(email="root@root.fr")
        cls.imgsrv = ImageServer.objects.get(url="http://server")

    def clear_caches(self):
        # Clean content type cache for SQL requests checks consistency
        ContentType.objects.clear_cache()

        # Clear the local cached properties so that it is re-fetched on each test
        # to avoid intermittently changing query counts.
        # Using `del` on a cached property that has not been accessed yet can cause an AttributeError.
        # If it wasn't accessed, then it wasn't fetched yet, so this doesn't matter.
        try:
            del ImageServer.objects.local
        except AttributeError:
            pass
        try:
            del WorkerVersion.objects.imports_version
        except AttributeError:
            pass
        try:
            del ImageServer.objects.ingest
        except AttributeError:
            pass

    def setUp(self):
        super().setUp()
        self.clear_caches()

    @contextmanager
    def subTest(self, *args, **kwargs):
        self.clear_caches()
        with super().subTest(*args, **kwargs):
            yield

    def assertExactQueries(
            self,
            path: Union[Path, str],
            func: Optional[Callable] = None,
            *args,
            using: str = DEFAULT_DB_ALIAS,
            params: Union[Iterable, Mapping[str, Any]] = [],
            skip: int = 0,
            **kwargs) -> Optional[_AssertExactQueriesContext]:
        """
        Assert that a function call causes exactly the SQL queries specified in a given file.
        When used as a context manager, asserts than any queries occurring while in the context match the given file.
        This function supports all of Django's `assertNumQueries` syntax:

        >>> self.assertExactQueries('path/to/file.sql', do_something)
        >>> with self.assertExactQueries('path/to/file.sql'):
        ...     do_something()

        Paths will be relative to the SQL_VALIDATION_DIR setting.
        SQL queries can often hold IDs or other dynamic values that depend on the test context.
        To avoid having to update IDs or rewrite the fixture file every time, you can use the `params` argument
        and give it anything supported by `str.format`; either an iterable or a mapping. For example:

        ```
        SELECT "documents_element"."name" FROM "documents_element" WHERE "documents_element"."id" = '{id}'::uuid;
        ```

        >>> with self.assertExactQueries('path/to/file.sql', params={'id': element_id}):
        ...     Element.objects.filter(id=element_id).values('name')

        To make first runs easier, when the expect SQL file does not exist, assertExactQueries will try to write
        the current SQL queries to it and warn you using an `AssertionError`. Make sure to check this new SQL file
        as subsequent runs of unit tests will no longer fail.

        In API endpoint tests, some queries do not really need to be tested and are annoying to mock, such as
        session queries that use the `django_session` table and are the first query to be made when you call an endpoint
        using the test client. You can skip those queries and go straight to the point using `skip`.
        This will skip the first query it finds:

        >>> with self.assertExactQueries('my_endpoint.sql', skip=1):
        ...     self.client.get(reverse('api:my-endpoint'))
        """
        context = _AssertExactQueriesContext(self, path, params, connections[using], skip=skip)

        if func is None:
            return context

        with context:
            func(*args, **kwargs)


class FixtureTestCase(FixtureMixin, TestCase):
    """
    Django test case with the database fixture
    """


class FixtureAPITestCase(FixtureMixin, APITestCase):
    """
    Django REST Framework test case with the database fixture
    """


@contextmanager
def force_constraints_immediate():
    """
    Force all unique, primary key, foreign key and exclusion constraints to be
    verified immediately, instead of being deferred until the end of a transaction
    as they are by default.

    This is useful to test code that is supposed to run outside of a transaction
    while in a Django TransactionTestCase or TestCase or an DRF APITestCase,
    as all of those classes will wrap each test into a transaction.

    https://www.postgresql.org/docs/current/sql-set-constraints.html
    """
    with connections[DEFAULT_DB_ALIAS].cursor() as cursor:
        cursor.execute("SET CONSTRAINTS ALL IMMEDIATE")
    yield
    with connections[DEFAULT_DB_ALIAS].cursor() as cursor:
        cursor.execute("SET CONSTRAINTS ALL DEFERRED")
