import logging
import urllib.parse

from django.conf import settings
from django.contrib.auth import login, logout
from django.contrib.auth.forms import PasswordResetForm
from django.contrib.auth.tokens import default_token_generator
from django.core.mail import send_mail
from django.template.loader import render_to_string
from django.urls import reverse
from django.utils.http import urlsafe_base64_encode
from django_rq.queues import get_queue
from django_rq.settings import QUEUES
from drf_spectacular.types import OpenApiTypes
from drf_spectacular.utils import OpenApiParameter, extend_schema, extend_schema_view
from rest_framework import serializers, status
from rest_framework.exceptions import AuthenticationFailed, NotFound, PermissionDenied, ValidationError
from rest_framework.generics import CreateAPIView, ListAPIView, RetrieveDestroyAPIView, RetrieveUpdateDestroyAPIView
from rest_framework.response import Response
from rest_framework.views import APIView
from rq.job import JobStatus

from arkindex.documents.models import Corpus
from arkindex.project.permissions import IsAuthenticatedOrReadOnly, IsVerified
from arkindex.users.models import Group, Role, Scope, User, UserScope
from arkindex.users.serializers import (
    EmailLoginSerializer,
    JobSerializer,
    NewUserSerializer,
    PasswordResetConfirmSerializer,
    PasswordResetSerializer,
    UserSerializer,
)

logger = logging.getLogger(__name__)

# Process tasks running in RQ are hidden from user jobs
VISIBLE_QUEUES = [
    q for q in QUEUES.keys()
    if q != "tasks"
]


@extend_schema(tags=["users"])
@extend_schema_view(
    get=extend_schema(description="Retrieve information about the authenticated user"),
    patch=extend_schema(description="Update a user's password. This action is not allowed to users without confirmed e-mails."),
    put=extend_schema(description="Update a user's password. This action is not allowed to users without confirmed e-mails."),
    delete=extend_schema(operation_id="Logout", description="Log out from the API")
)
class UserRetrieve(RetrieveUpdateDestroyAPIView):
    """
    Retrieve information about the authenticated user
    """
    permission_classes = (IsAuthenticatedOrReadOnly, )
    serializer_class = UserSerializer

    def get_object(self):
        return self.request.user

    def retrieve(self, *args, **kwargs):
        if not self.request.user.is_authenticated:
            # Include features on GET to let the frontend still know the instance features when not logged in
            return Response(
                {
                    "detail": PermissionDenied.default_detail,
                    "features": settings.ARKINDEX_FEATURES,
                },
                status=status.HTTP_403_FORBIDDEN,
            )

        return super().retrieve(*args, **kwargs)

    def perform_destroy(self, instance):
        logout(self.request)


@extend_schema_view(post=extend_schema(
    operation_id="Register",
    responses={201: UserSerializer},
    tags=["users"],
))
class UserCreate(CreateAPIView):
    """
    Register as a new user.
    """
    serializer_class = NewUserSerializer

    def create(self, request, *args, **kwargs):
        if not settings.ARKINDEX_FEATURES["signup"]:
            raise ValidationError(["Registration has been disabled on this instance."])

        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        if User.objects.filter(email=serializer.validated_data["email"]).exists():
            raise ValidationError({"email": ["An user with this email address already exists."]})

        user = serializer.save()
        login(self.request, user)

        activation_url = "{}?{}".format(
            self.request.build_absolute_uri(reverse("api:user-token")),
            urllib.parse.urlencode({
                "email": user.email,
                "token": default_token_generator.make_token(user),
            }),
        )
        sent = send_mail(
            # Subject cannot have any newlines
            subject="".join(render_to_string("registration/verification_email_subject.txt").splitlines()),
            message=render_to_string(
                "registration/verification_email.html",
                context={
                    "url": activation_url,
                },
                request=self.request,
            ),
            from_email=None,
            recipient_list=[user.email],
            fail_silently=True,
        )
        if sent == 0:
            logger.error(f"Failed to send registration email to {user.email}")

        return Response(UserSerializer(user).data, status=status.HTTP_201_CREATED)


@extend_schema_view(post=extend_schema(
    operation_id="Login",
    responses={201: UserSerializer},
    tags=["users"]
))
class UserEmailLogin(CreateAPIView):
    """
    Login using an email and a password
    """
    serializer_class = EmailLoginSerializer

    def create(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        try:
            user = User.objects.get_by_natural_key(serializer.validated_data["email"])
        except User.DoesNotExist:
            raise AuthenticationFailed

        if not user.has_usable_password() or not user.check_password(serializer.validated_data["password"]):
            raise AuthenticationFailed

        login(self.request, user)

        headers = self.get_success_headers(serializer.data)
        return Response(
            UserSerializer(user).data,
            status=status.HTTP_201_CREATED,
            headers=headers,
        )


class UserEmailVerification(APIView):
    """
    Verify a user's email address
    """

    @extend_schema(
        operation_id="VerifyEmail",
        tags=["users"],
        parameters=[
            OpenApiParameter(
                "email",
                description="E-mail to verify",
                type=OpenApiTypes.EMAIL,
                required=True,
            ),
            OpenApiParameter(
                "token",
                description="Verification token",
                required=True,
            )
        ],
        responses={302: serializers.Serializer}
    )
    def get(self, *args, **kwargs):
        if not all(arg in self.request.GET for arg in ("email", "token")):
            raise ValidationError

        try:
            user = User.objects.get_by_natural_key(self.request.GET["email"])
        except User.DoesNotExist:
            raise AuthenticationFailed

        if not default_token_generator.check_token(user, self.request.GET["token"]):
            raise AuthenticationFailed

        user.verified_email = True
        user.save()
        UserScope.objects.bulk_create([
            UserScope(user=user, scope=Scope.CreateIIIFImage),
            UserScope(user=user, scope=Scope.UploadS3Image),
        ])

        if user.has_usable_password():
            login(self.request, user)
            # A user with a verified email is granted a private trial corpus with admin rights
            if not Corpus.objects.admin(user=user).exists():
                corpus = Corpus.objects.create(
                    name="My Project",
                    description=f"Project for {user.email}"
                )
                corpus.memberships.create(
                    user=user,
                    level=Role.Admin.value
                )
                # Add default types
                corpus.create_default_types()

            # Redirtect to home page
            url = "/"
        else:
            url = reverse("password_reset_confirm", kwargs={
                "uidb64": urlsafe_base64_encode(str(user.id).encode()),
                "token": self.request.GET["token"],
            })

        # If a default group into which to add newly registered and verified users has been
        # defined, then add the user to that group
        if settings.SIGNUP_DEFAULT_GROUP:
            Group.objects.get(pk=settings.SIGNUP_DEFAULT_GROUP).memberships.create(user=user, level=Role.Guest.value)

        return Response(
            status=status.HTTP_302_FOUND,
            headers={"Location": url},
        )


@extend_schema_view(post=extend_schema(operation_id="ResetPassword", tags=["users"]))
class PasswordReset(CreateAPIView):
    """
    Start a password reset flow
    """
    serializer_class = PasswordResetSerializer

    def create(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        form = PasswordResetForm(serializer.validated_data)

        if form.is_valid():
            form.save(
                request=self.request,
                use_https=self.request.is_secure(),
                token_generator=default_token_generator,
                email_template_name="registration/password_reset_email.html",
                subject_template_name="registration/password_reset_subject.txt",
            )
            headers = self.get_success_headers(serializer.data)
            return Response(
                status=status.HTTP_201_CREATED,
                headers=headers,
            )
        return Response(
            form.errors.get_json_data(),
            status=status.HTTP_400_BAD_REQUEST,
            headers=self.default_response_headers
        )


@extend_schema_view(post=extend_schema(operation_id="PasswordResetConfirm", tags=["users"]))
class PasswordResetConfirm(CreateAPIView):
    """
    Confirm a password reset using data from the confirmation e-mail
    """
    serializer_class = PasswordResetConfirmSerializer


@extend_schema(tags=["jobs"])
class JobList(ListAPIView):
    """
    List asynchronous jobs linked to the current user.
    """
    permission_classes = (IsVerified, )
    serializer_class = JobSerializer
    pagination_class = None

    def get_queryset(self):
        if not self.request:
            return []
        return self.request.user.get_rq_jobs()


@extend_schema(
    tags=["jobs"],
    parameters=[
        OpenApiParameter(
            "id",
            location=OpenApiParameter.PATH,
            type=OpenApiTypes.UUID,
        )
    ]
)
class JobRetrieve(RetrieveDestroyAPIView):
    """
    Retrieve a single job by ID.
    """
    permission_classes = (IsVerified, )
    serializer_class = JobSerializer

    def get_object(self):
        for queue_name in VISIBLE_QUEUES:
            job = get_queue(queue_name).fetch_job(str(self.kwargs["pk"]))
            if not job:
                continue
            if job.user_id != self.request.user.id:
                raise NotFound
            return job
        raise NotFound

    def perform_destroy(self, instance):
        if instance.get_status(refresh=False) == JobStatus.STARTED:
            raise ValidationError(["Cannot delete a running job."])
        instance.delete()
